import numpy as np
import librosa
import random

class Signal():

    def __init__(self, signal, sr):

        self.full_signal    = signal
        self.sr             = sr

        # at first self.signal stores the full signal but it gets overwritten
        self.signal         = signal

        self.n_fft          =   1024
        self.frame_length   =   1024
        self.hop_length     =   512
        self.number_of_mfccs =  40
        self.f_min          =   20

        self.n = None

    def extract_spectral_features(self):

        chroma_stft         = self._extract_chroma_stft()
        melspectrogram      = self._extract_melspectrogram()
        mfcc                = self._extract_mfcc()
        rms                 = self._extract_rms()
        spectral_centroid   = self._extract_spectral_centroid()
        spectral_bandwidth  = self._extract_spectral_bandwidth()
        spectal_contrast    = self._extract_spectral_contrast()
        spectral_flatness   = self._extract_spectral_flatness()
        spectral_rolloff    = self._extract_spectral_rolloff()
        poly_features       = self._extract_poly_features()
        zero_crossing_rate  = self._extract_zero_crossing_rate()
        
        spectral_features = [chroma_stft, melspectrogram, mfcc, rms, spectral_centroid, 
                            spectral_bandwidth, spectal_contrast, spectral_flatness, spectral_rolloff,
                            poly_features, zero_crossing_rate]

        return spectral_features

    def extract_spectral_features_from_x_windows(self, number_of_windows = 100):

        chroma_stft         = self._extract_chroma_stft()
        chroma_stft         = self._extract_features_from_random_window_range(chroma_stft, number_of_windows)

        melspectrogram      = self._extract_melspectrogram()
        melspectrogram      = self._extract_features_from_random_window_range(melspectrogram, number_of_windows)

        mfcc                = self._extract_mfcc()
        mfcc                = self._extract_features_from_random_window_range(mfcc, number_of_windows)

        rms                 = self._extract_rms()
        rms                 = self._extract_features_from_random_window_range(rms, number_of_windows)

        spectral_centroid   = self._extract_spectral_centroid()
        spectral_centroid   = self._extract_features_from_random_window_range(spectral_centroid, number_of_windows)

        spectral_bandwidth  = self._extract_spectral_bandwidth()
        spectral_bandwidth  = self._extract_features_from_random_window_range(spectral_bandwidth, number_of_windows)

        spectal_contrast    = self._extract_spectral_contrast()
        spectal_contrast    = self._extract_features_from_random_window_range(spectal_contrast, number_of_windows)

        spectral_flatness   = self._extract_spectral_flatness()
        spectral_flatness   = self._extract_features_from_random_window_range(spectral_flatness, number_of_windows)

        spectral_rolloff    = self._extract_spectral_rolloff()
        spectral_rolloff    = self._extract_features_from_random_window_range(spectral_rolloff, number_of_windows)

        poly_features       = self._extract_poly_features()
        poly_features       = self._extract_features_from_random_window_range(poly_features, number_of_windows)

        zero_crossing_rate  = self._extract_zero_crossing_rate()  
        zero_crossing_rate  = self._extract_features_from_random_window_range(zero_crossing_rate, number_of_windows)


        spectral_features_from_x_windows = [chroma_stft, melspectrogram, mfcc, rms, spectral_centroid, 
                                            spectral_bandwidth, spectal_contrast, spectral_flatness, spectral_rolloff,
                                            poly_features, zero_crossing_rate]

        return spectral_features_from_x_windows
    
    def extract_means_of_spectral_features(self):

        chroma_stft         = self._extract_chroma_stft()
        chroma_stft         = self._calculate_feature_average(chroma_stft)

        melspectrogram      = self._extract_melspectrogram()
        melspectrogram      = self._calculate_feature_average(melspectrogram)

        mfcc                = self._extract_mfcc()
        mfcc                = self._calculate_feature_average(mfcc)

        rms                 = self._extract_rms()
        rms                 = self._calculate_feature_average(rms)

        spectral_centroid   = self._extract_spectral_centroid()
        spectral_centroid   = self._calculate_feature_average(spectral_centroid)

        spectral_bandwidth  = self._extract_spectral_bandwidth()
        spectral_bandwidth  = self._calculate_feature_average(spectral_bandwidth)

        spectal_contrast    = self._extract_spectral_contrast()
        spectal_contrast    = self._calculate_feature_average(spectal_contrast)

        spectral_flatness   = self._extract_spectral_flatness()
        spectral_flatness   = self._calculate_feature_average(spectral_flatness)

        spectral_rolloff    = self._extract_spectral_rolloff()
        spectral_rolloff    = self._calculate_feature_average(spectral_rolloff)

        poly_features       = self._extract_poly_features()
        poly_features       = self._calculate_feature_average(poly_features)

        zero_crossing_rate  = self._extract_zero_crossing_rate()  
        zero_crossing_rate  = self._calculate_feature_average(zero_crossing_rate)


        means_of_spectral_features = [chroma_stft, melspectrogram, mfcc, rms, spectral_centroid, 
                                    spectral_bandwidth, spectal_contrast, spectral_flatness, spectral_rolloff,
                                    poly_features, zero_crossing_rate]

        return means_of_spectral_features

    def _moving_average(self, x, N):

        x_ave = np.convolve(x, np.ones(N)/N, mode = "valid")

        return x_ave

    def _extract_chroma_stft(self):

        chroma_stft_features = librosa.feature.chroma_stft(y = self.signal, sr = self.sr, 
                                                            n_fft = self.n_fft, hop_length = self.hop_length)                         

        self.n = 12
        
        return chroma_stft_features

    def _extract_melspectrogram(self):

        melspectrogram_features = librosa.feature.melspectrogram(y = self.signal, sr = self.sr, 
                                                                n_fft = self.n_fft, hop_length = self.hop_length)

        self.n = 128
        
        return melspectrogram_features

    def _extract_mfcc(self):

        mfcc_features = librosa.feature.mfcc(y = self.signal, sr = self.sr, n_mfcc = self.number_of_mfccs)

        self.n = self.number_of_mfccs
        
        return mfcc_features

    def _extract_rms(self):

        rms_features = librosa.feature.rms(y = self.signal, 
                                            frame_length = self.frame_length, hop_length = self.hop_length)

        self.n = 1
        
        return rms_features

    def _extract_spectral_centroid(self):

        spectral_centroid_features = librosa.feature.spectral_centroid(y = self.signal, sr = self.sr, 
                                                                    n_fft = self.n_fft, hop_length = self.hop_length)

        self.n = 1
        
        return spectral_centroid_features

    def _extract_spectral_bandwidth(self):

        spectral_bandwidth_features = librosa.feature.spectral_bandwidth(y = self.signal, sr = self.sr, 
                                                                n_fft = self.n_fft, hop_length = self.hop_length)

        self.n = 1
        
        return spectral_bandwidth_features

    def _extract_spectral_contrast(self):

        spectral_contrast_features = librosa.feature.spectral_contrast(y = self.signal, sr = self.sr, 
                                                                n_fft = self.n_fft, hop_length = self.hop_length, fmin = self.f_min)

        self.n = 7
        
        return spectral_contrast_features

    def _extract_spectral_flatness(self):

        spectral_flatness_features = librosa.feature.spectral_flatness(y = self.signal, 
                                                                n_fft = self.n_fft, hop_length = self.hop_length)

        self.n = 1
        
        return spectral_flatness_features

    def _extract_spectral_rolloff(self):

        spectral_rolloff_features = librosa.feature.spectral_rolloff(y = self.signal, sr = self.sr, 
                                                                n_fft = self.n_fft, hop_length = self.hop_length)

        self.n = 1
        
        return spectral_rolloff_features

    def _extract_poly_features(self):

        poly_features = librosa.feature.poly_features(y = self.signal, sr = self.sr, 
                                                    n_fft = self.n_fft, hop_length = self.hop_length)

        self.n = 2
        
        return poly_features

    def _extract_zero_crossing_rate(self):

        zero_crossing_rate_features = librosa.feature.zero_crossing_rate(y = self.signal,
                                                                    frame_length = self.frame_length, hop_length = self.hop_length)

        self.n = 1
        
        return zero_crossing_rate_features 

    def _calculate_feature_average(feature):

        feature_average = np.sum(feature, axis=1)

        return feature_average

    def _extract_features_from_random_window_range(self, feature, number_of_windows = 100):

        # n is number of features per window, unique to each feature
        
        number_of_columns = feature.shape[1]

        if number_of_columns > number_of_windows:

            starting_point_edge = number_of_columns - number_of_windows
            starting_point = random.randint(0, starting_point_edge)

            features_from_random_window_range = feature[:, starting_point:starting_point + number_of_windows]

        elif number_of_columns < number_of_windows:

            columns_of_padding_required = number_of_windows - number_of_columns
            zeros_padding = np.zeros((self.n, columns_of_padding_required), dtype = int)

            features_from_random_window_range = np.concatenate((feature, zeros_padding), axis = 1)

        elif number_of_columns == number_of_windows:

            features_from_random_window_range = feature        
        
        return features_from_random_window_range